/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <X11/Xproto.h>
#include "debug.h"


static int debug = 0;

int ConnectSocketToDisplay(display_var)
char *display_var;
{
   int                     socket_to_display;
   int                     port_to_display;
   char                    host_name[20];
   struct sockaddr_in      host_addr;
   int                     display_number;
   struct hostent          *host_info;


   DEBUG("DISPLAY = %s\n",display_var);
   if ( ( socket_to_display = socket( AF_INET, SOCK_STREAM, 0 ))
        < 0 )  {
      return (-1);
   }
   DisplayNameAnalyze(display_var,host_name,&display_number);
   port_to_display = X_TCP_PORT + display_number;
   host_info =  gethostbyname(host_name);
   if  ( host_info  != NULL ) {
        bcopy(host_info->h_addr, &(host_addr.sin_addr),
              host_info->h_length);
   }
   else {
        host_addr.sin_addr.s_addr = inet_addr(host_name);
        if ((host_info = gethostbyaddr(&host_addr.sin_addr.s_addr,
             sizeof(host_addr.sin_addr.s_addr),AF_INET))== NULL) {
/*
           printf( "ConnectTo(): Can't find host \"%s\"\n", host_name);
           return (-1);
*/
        }
   }
   host_addr.sin_family = AF_INET;
   host_addr.sin_port = htons(port_to_display);
   if (connect( socket_to_display, &host_addr, sizeof(host_addr) ) != 0) {
       close(socket_to_display);
       printf( "Unable to connect to %s port =%d, ntohs(port)=%d\n", host_name,
               port_to_display,ntohs(port_to_display));
       return (-1);
   }
   DEBUG("Connection to the X-server succeed!\n");
   return socket_to_display;
}



DisplayNameAnalyze(display_var,host_name,display_number)
   char *display_var;
   char *host_name;
   int  *display_number;
{
   char                    *index1,
                           *index2,
                           *index();

   strcpy(host_name,display_var);
   if ( ( index1 = index(host_name,':')) == NULL) {
      printf("ERROR: Can't find ':' in DISPLAY variable %s\n",
             host_name);
      exit(0);
   }
   index2 = index(display_var,".");
   *index1 = '\0';

   index1++;

   if (index2 != NULL) {
      index2 = '\0';
   }
   sscanf(index1,"%d",display_number);
   DEBUG("host_name = %s\n",host_name);
   DEBUG("display_number = %d\n",*display_number);
}

int OpenPublicDisplaySocketAtLocalHost(new_display_number)
   int *new_display_number;
{
   int public_socket;

   *new_display_number = 0;

   for (public_socket=(-1); public_socket<0; ) {
      (*new_display_number)++;
      public_socket =
           OpenPublicPortedSocketAtLocalHost(htons(X_TCP_PORT+
                                                   (*new_display_number)));
   }  
   return public_socket;
}

int OpenPublicPortedSocketAtLocalHost( port )
int   port;
{
   int                public_socket;
   struct sockaddr_in public_socket_addr;

   if ((public_socket = socket( AF_INET, SOCK_STREAM, 0)) < 0) {
                return( -1 );
   }

   public_socket_addr.sin_family = AF_INET;
   public_socket_addr.sin_addr.s_addr = INADDR_ANY;
   public_socket_addr.sin_port = port;
   if ( bind(public_socket,&public_socket_addr,sizeof(public_socket_addr))
        != 0 ) {
           close( public_socket );
           return( -1 );
   }

   if (listen( public_socket , 5) < 0 ) {
         return( -1 );
   }
   return(public_socket);
}

char *GetNewDisplayVariable(new_display_number)
   int new_display_number;
{
   char *new_display_variable;
   int  new_display_variable_length;

   new_display_variable = (char *) malloc(32*sizeof(char));
   gethostname(new_display_variable,30);
   DEBUG("   new_display_variable = %s\n",new_display_variable);
   new_display_variable_length = strlen(new_display_variable);
   new_display_variable[new_display_variable_length] = ':';
   new_display_variable_length++;

   sprintf(new_display_variable+new_display_variable_length,
           "%d",new_display_number);
   DEBUG("new_display_variable = %s\n",new_display_variable);
   return new_display_variable;
}

int ReadBytes(socket_to_read,buffer,number_of_bytes_to_read)
   int socket_to_read;
   BYTE *buffer;
   int number_of_bytes_to_read;
{
   int number_of_bytes_just_read;
   int debug = 0;

   while ( number_of_bytes_to_read >0 ) {
      number_of_bytes_just_read = read(socket_to_read,buffer,
                                       number_of_bytes_to_read);
      if (number_of_bytes_just_read > 0) {
         number_of_bytes_to_read -= number_of_bytes_just_read;
         buffer += number_of_bytes_just_read;
      }
   }
}


SendBytes(socket_to_send,buffer,number_of_bytes_to_send)
   int  socket_to_send;
   BYTE *buffer;
   int  number_of_bytes_to_send;
{
   int number_of_bytes_just_sent;
   int debug = 0;


   while (number_of_bytes_to_send > 0) {
      number_of_bytes_just_sent = write(socket_to_send,buffer,  
                                        number_of_bytes_to_send);
      if (number_of_bytes_just_sent>0) {
         number_of_bytes_to_send -= number_of_bytes_just_sent;
         buffer += number_of_bytes_just_sent;
      }
   }
}


int AcceptSocketConnection(public_socket)
   int public_socket;
{
   struct sockaddr_in peer_address;
   int                address_length;
   int                real_socket;


   address_length = sizeof(struct sockaddr_in);
   real_socket = accept(public_socket,
                        (struct sockaddr *) &peer_address,
                        &address_length);

   if (real_socket<0) {
      printf("Connection Error in accepting connection.\n");
   }

   DEBUG("Accepting Succeed!\n");
   return real_socket;
}


