/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */

#define  NEED_REPLIES           /* Tell the header to include the replies */
#define    NEED_EVENTS

#include <X11/Xproto.h>
#include <X11/Xlib.h>
#include "vs.h"
#include "debug.h"


static debug = 0;

ReceiveSetupRequest(socket_to_receive,setup_request)
   int                socket_to_receive;
   BYTE               *setup_request;
{
   xConnClientPrefix  *setup_request_prefix;
   int                LengthOfSetupRequest();

   ReadBytes(socket_to_receive,setup_request,sz_xConnClientPrefix);
   ReadBytes(socket_to_receive,setup_request+sz_xConnClientPrefix,
             LengthOfSetupRequest(setup_request) - sz_xConnClientPrefix);
}

SendSetupRequest(socket_to_send,setup_request)
   int socket_to_send;
   BYTE *setup_request;
{
   int                LengthOfSetupRequest();

   SendBytes(socket_to_send,setup_request,LengthOfSetupRequest(setup_request));
}

int LengthOfSetupRequest(setup_request)
   BYTE               *setup_request;
{
   xConnClientPrefix  *setup_request_prefix;
   int                additional_size;

   setup_request_prefix = (xConnClientPrefix  *) setup_request;
   additional_size =
            AddPadBy4( (int) (setup_request_prefix->nbytesAuthProto)  ) +
            AddPadBy4( (int) (setup_request_prefix->nbytesAuthString) ) ;

   return (sz_xConnClientPrefix + additional_size);
}

ReceiveSetupReply(socket_to_receive,setup_reply)
   int socket_to_receive;
   BYTE *setup_reply;
{
   int  SetupReplyLength();
   BYTE *reply_with_sucess;
   int  debug = 0;

   DEBUG("  start receiving setup reply head.\n");
   ReadBytes(socket_to_receive,setup_reply,sz_xConnSetupPrefix);
   DEBUG("  end   receiving setup reply head.\n");
   reply_with_sucess = (BYTE *) setup_reply;
   if ( (*reply_with_sucess) != 1) {
      printf("  Setup failure from X-server! : %d\n",(*reply_with_sucess));
      /*
      exit(0);
      */
   }
   DEBUG("  start receiving setup reply tail.\n");
   ReadBytes(socket_to_receive,setup_reply+sz_xConnSetupPrefix,
             SetupReplyLength(setup_reply)-sz_xConnSetupPrefix);
   DEBUG("  end   receiving setup reply tail.\n");
}

int SetupReplyLength(setup_reply)
   BYTE *setup_reply;
{
   xConnSetupPrefix        *set_up_reply_prefix;

   set_up_reply_prefix = (xConnSetupPrefix *) setup_reply;
   if (set_up_reply_prefix->success  == 0)   {
      return(sz_xConnSetupPrefix+(int)(set_up_reply_prefix->lengthReason));
   }
   else {
      return (sz_xConnSetupPrefix + 4*(int)(set_up_reply_prefix->length));
   }
}



Window GetRootWindowInSetupReply(setup_reply)
   BYTE *setup_reply;
{
   Window *root_window;
   BYTE   *FindByteLocationOfRootWindowInSetupReply();

   root_window=(Window *)FindByteLocationOfRootWindowInSetupReply(setup_reply);
   return (*root_window);
}

CARD32 BlackPixelOfDefaultScreen(setup_reply)
{
   xWindowRoot *root_screen;

   root_screen = (xWindowRoot *) 
                         FindByteLocationOfRootWindowInSetupReply(setup_reply);
   return (root_screen->blackPixel);
}

Colormap DefaultColormapOfDefaultScreen(setup_reply)
{
   xWindowRoot *root_screen;
   root_screen = (xWindowRoot *)
                         FindByteLocationOfRootWindowInSetupReply(setup_reply);
   return (root_screen->defaultColormap);
}


CARD32 WhitePixelOfDefaultScreen(setup_reply)
{
   xWindowRoot *root_screen;

   root_screen = (xWindowRoot *) 
                        FindByteLocationOfRootWindowInSetupReply(setup_reply);
   return (root_screen->whitePixel);
}

CARD32 OriginalInputMaskOfRootWindow(setup_reply)
{
   xWindowRoot *root_screen;

   root_screen = (xWindowRoot *)
                        FindByteLocationOfRootWindowInSetupReply(setup_reply);
   return (root_screen->currentInputMask);
}

VisualID RootVisualID(setup_reply)
{
   xWindowRoot *root_screen;

   root_screen = (xWindowRoot *)
                        FindByteLocationOfRootWindowInSetupReply(setup_reply);
   return (root_screen->rootVisualID);
}


BYTE *FindByteLocationOfRootWindowInSetupReply(setup_reply)
   BYTE *setup_reply;
{
   xConnSetupPrefix        *set_up_reply_prefix;
   xConnSetup              *set_up_reply_body;
   BYTE                    *vendor;
   xPixmapFormat           *pixmap_formats;
   BYTE                    *root_screen;
   int                     vendor_size,
                           root_area_size,
                           max_request_size;
   int                     number_of_roots;


   set_up_reply_prefix = (xConnSetupPrefix *) setup_reply;
   if (set_up_reply_prefix->success  == 0)   {
      printf("X-server replied that the setup process is incorrect!\n");
      exit(0);
   }

   set_up_reply_body = (xConnSetup *) (setup_reply + sz_xConnSetupPrefix);
   max_request_size = (int) (set_up_reply_body->maxRequestSize);
   DEBUG("max_request_size = %d\n",max_request_size);
/*
   if (max_request_size*4 > MAXIMUM_SIZE_OF_A_PROTOCOL) {
      printf("Warning : the possible X-request is %d bytes.",
             max_request_size*4);
      printf("  But the program allocates only %d bytes.\n",
             MAXIMUM_SIZE_OF_A_PROTOCOL);
      printf("          Change MAXIMUM_SIZE_OF_A_PROTOCOL in common.h!\n");
   }
*/
   number_of_roots = (set_up_reply_body->numRoots);
   if (number_of_roots != 1) {
      printf("Warning : there are %d roots which was supposed to be 1 ",
             number_of_roots);
      printf("when developing this program!\n");
   }
   vendor = (BYTE *) set_up_reply_body + sz_xConnSetup;
   vendor_size = (int) set_up_reply_body->nbytesVendor;

   pixmap_formats = (xPixmapFormat*) (vendor + AddPadBy4(vendor_size));

   root_area_size =  (int)(set_up_reply_prefix->length) * 4 
                     - sz_xConnSetup - AddPadBy4(vendor_size) -
                     (int)(set_up_reply_body->numFormats * sz_xPixmapFormat) ;
   root_screen = (BYTE *) pixmap_formats +
                 set_up_reply_body->numFormats * sz_xPixmapFormat;
   return (BYTE *) &(((xWindowRoot *)root_screen)->windowId) ;
}


SendSetupReply(socket_to_send,setup_reply)
   int socket_to_send;
   BYTE *setup_reply;
{
   int SetupReplyLength();

   SendBytes(socket_to_send,setup_reply,SetupReplyLength(setup_reply) );
}

CARD32 ResourceMaskFromSetupReply(setup_reply) 
   BYTE *setup_reply;
{
   xConnSetup              *set_up_reply_body;

   set_up_reply_body = (xConnSetup *) (setup_reply + sz_xConnSetupPrefix);
   return (set_up_reply_body->ridMask);
}
   
CARD32 ResourceBaseFromSetupReply(setup_reply)
   BYTE *setup_reply;
{
   xConnSetup              *set_up_reply_body;

   set_up_reply_body = (xConnSetup *) (setup_reply + sz_xConnSetupPrefix);
   return (set_up_reply_body->ridBase);
}

int MaxRequestSizeFromSetupReply(setup_reply)
   BYTE *setup_reply;
{
   xConnSetup              *set_up_reply_body;

   set_up_reply_body = (xConnSetup *) (setup_reply + sz_xConnSetupPrefix);
   return ( (int) (set_up_reply_body->maxRequestSize) );
}



   

BOOL IsAReply(reply_or_event) 
   BYTE           *reply_or_event;
{
    if (*reply_or_event == 1) return True;
    else return False;
}
    


int ReceiveReplyOrEvent(socket_to_receive,reply_or_event)
   int            socket_to_receive;
   BYTE           *reply_or_event;
{
   int            additional_reply_length;
   xGenericReply  *x_generic_reply;
   BYTE           *reply,
                  *event;
   int            tested_result;
   int            debug = 0;

   tested_result = read(socket_to_receive,reply_or_event,1);
   if (tested_result<=0) {
      if (tested_result>=0) {DEBUG("tested_result = %d\n",tested_result);}
      return tested_result;
   }
   if (*reply_or_event == 1) {
      DEBUG("  A reply is sent.\n");
      reply = reply_or_event;
      ReadBytes(socket_to_receive,reply+1,sz_xGenericReply-1);
      x_generic_reply = (xGenericReply *) reply;

      additional_reply_length =
                     ( (int ) (x_generic_reply->length) ) * 4 ;
      DEBUG("additional_reply_length = %d\n",additional_reply_length);
      ReadBytes(socket_to_receive,reply+sz_xGenericReply,
                additional_reply_length);
      return (sz_xGenericReply+additional_reply_length);
   }
   else {
      event = reply_or_event;
      ReadBytes(socket_to_receive,event+1,sz_xEvent-1);
      return sz_xEvent;
   }
}


SendReplyOrEvent(socket_to_send,reply_or_event)
   int socket_to_send;
   BYTE *reply_or_event;
{
   if (*reply_or_event== 1) {
      SendReply(socket_to_send,reply_or_event);
   }
   else {
      SendEvent(socket_to_send,reply_or_event);
   }
}


SendReply(socket_to_send,reply)   
   int  socket_to_send;
   BYTE *reply;
{
   SendBytes(socket_to_send,reply,ReplyLength(reply));
}

int ReplyLength(reply)
   BYTE *reply;
{
   xGenericReply  *x_generic_reply;

   x_generic_reply = (xGenericReply *) (reply);
   return ( sz_xGenericReply + ( (int ) (x_generic_reply->length) ) * 4 );
}

SendEvent(socket_to_send,event)
   int  socket_to_send;
   BYTE *event;
{
   SendBytes(socket_to_send,event,sz_xEvent);
}
   
ReceiveRequestProtocol(socket_to_receive,request_protocol,request_type)
   int  socket_to_receive;
   BYTE *request_protocol;
   int  *request_type;
{
   int RequestLength(); 

   if ( read(socket_to_receive,request_protocol,sz_xReq) <= 0 ) return 0;
   ReadBytes(socket_to_receive,request_protocol+sz_xReq,
             RequestLength(request_protocol) - sz_xReq);
   return (RequestLength(request_protocol));
}


SendRequestProtocol(socket_to_send,request_protocol)
   int  socket_to_send;
   BYTE *request_protocol;
{
   SendBytes(socket_to_send,request_protocol,RequestLength(request_protocol));
}



int RequestLength(request_protocol) 
   BYTE *request_protocol;
{
    xReq *x_request_head;

    x_request_head = (xReq *) request_protocol;

    return ( ((int) (x_request_head->length)) * 4 );
}

       

int RequestType(request_protocol)
   BYTE *request_protocol;
{
   xReq *x_request;
   int  request_type;

   x_request = (xReq *) request_protocol;

   request_type = (int) (x_request->reqType);
   return request_type;
}

BOOL ConfigureWindowUpDown(protocol)
   BYTE                *protocol;
{
   xConfigureWindowReq *configure_request;

   configure_request = (xConfigureWindowReq *) protocol;

   if (configure_request->reqType == X_ConfigureWindow) {        
      if (configure_request->mask & CWStackMode) { 
         return True;
      }
   }
   return False;
}

BOOL RedirectAnotherWindow(protocol)
   BYTE                *protocol;
{
   xChangeWindowAttributesReq *change_attribute_request;

   change_attribute_request = (xChangeWindowAttributesReq *) protocol;

   if ( change_attribute_request->reqType == X_ChangeWindowAttributes) {
      if (change_attribute_request->valueMask & CWOverrideRedirect ) {
         return True;
      }
   }
   return False;
}
                                      
Describe_X_PutImage_Request(protocol)
   BYTE                *protocol;
{
   xPutImageReq        *put_image_request;
   BYTE                *data;
   int                 data_size;

   put_image_request = (xPutImageReq *) protocol;
   data = (BYTE *) (protocol + sz_xPutImageReq);

   if (put_image_request->reqType != X_PutImage) {
      printf("   This is NOT X_PutImage request! \n");
      return;
   }
   printf(" -----------------  X_PutImage  -----------------\n");
   printf("   format: ");
   switch (put_image_request->format) {
      case XYBitmap: printf("XYBitmap. "); break;
      case XYPixmap: printf("XYPixmap. "); break;
      case ZPixmap : printf("ZPixmap.  ");  break;
   }
   printf("request length = %d bytes\n",put_image_request->length*4);
   data_size = put_image_request->length*4 - sz_xPutImageReq;
   printf("   ( %d bytes for images including pad)\n",data_size);

   printf("   drawable = 0x%08x, gc = 0x%08x\n",
          put_image_request->drawable,put_image_request->gc);
   printf("   width    = %3d, height = %3d\n",
          put_image_request->width,put_image_request->height);
   printf("   dstX     = %3d, dstY    = %3d\n",
          put_image_request->dstX,put_image_request->dstY);
   printf("   left_pad = %3d, depth   = %3d\n",
          put_image_request->leftPad,put_image_request->depth);
}





DescribeConfigureWindow(protocol)
   BYTE                *protocol;
{
   xConfigureWindowReq *configure_request;
   Window              *handled_window,*sibling_window;
   CARD16              *value_mask;
   unsigned long       *value_list_point;
   INT16               *location;
   CARD16              *length;
   BYTE                *stack_mode;

   configure_request = (xConfigureWindowReq *) protocol;
   if ( configure_request->reqType != X_ConfigureWindow ) {
      printf("Error: Trying to print with ConfigureWindow format but not");
      printf("match\n");
      return;
   }
   
   value_list_point = (unsigned long *) 
                                 (protocol + sz_xConfigureWindowReq);

   printf("--------------- ConfigureWindow ---------------\n");
   printf("       window = 0x%08x\n",configure_request->window);
   printf("   value mask = 0x%04x\n",configure_request->mask );

   if ( configure_request->mask & CWX ) {
      location = (INT16   *) value_list_point;
      printf( "       x  = %d\n" , (*location) );
      value_list_point++;
   }

   if ( configure_request->mask & CWY )  {
      location = (INT16   *) value_list_point;
      printf( "       y  = %d\n" , (*location) );
      value_list_point++;
   }

   if ( configure_request->mask & CWWidth )  {
      length = (CARD16 *) value_list_point;
      printf( "   width  = %d\n" , (*length) );
      value_list_point++;
      value_list_point++;
   }

   if ( configure_request->mask & CWHeight)  {
      length = (CARD16 *) value_list_point;
      printf( "  height  = %d\n" , (*length) );
      value_list_point++;
   }

   if ( configure_request->mask & CWBorderWidth)  {
      length = (CARD16 *) value_list_point;
      printf( "  border  = %d\n" , (*length) );
      value_list_point++;
   }


   if ( configure_request->mask & CWSibling)  {
      sibling_window = (Window *) value_list_point;
      printf( " sibling  = 0x%08x\n" , (*sibling_window) );
      value_list_point++;
   }

   if ( configure_request->mask & CWStackMode)  {
      printf(" stack    = ");
      stack_mode = (BYTE *) value_list_point;
      switch (*stack_mode) {
         case 0 : printf("Above"); break;
         case 1 : printf("Below"); break;
         case 2 : printf("TopIf"); break;
         case 3 : printf("ButtomIf"); break;
         case 4 : printf("Opposite"); break;
         default : printf("Error"); break;
     }
     printf("\n");
   }
   printf("-----------------------------------------------\n");
}

int AddPadBy4(a_number)
   int a_number;
{
   if (a_number%4 == 0) {
      return a_number;
   }
   else {
      return ( (a_number/4 + 1) * 4 );
   }
}


  
/*
   
SendCloseNoticeOfVirtualScreen(socket_to_send) 
   int socket_to_send;
{
   xReq   *x_request_head;
   static BOOL first_time = True;

   static BYTE close_request[] = {XCloseDisplayRequest, XCloseDisplayRequest,
                                 0,                    0                  };
   
   if (first_time == True) {
      x_request_head = (xReq *) close_request;
      x_request_head->length = 1;
      first_time = False;
   }
   SendBytes(socket_to_send,close_request,4);
}
*/
