/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */

/*
**  packet.c
**  Packet handlers
**
**
**  Modifications :
**                  FSH 9-5-91  Rewrote  HandleIncomingClientPacket().
**                  FSH 12-7-91 Changed SwapInt() and SwapLong() to macros
**                              in all the code  
*/


#include <stdio.h>
#include <stdlib.h>
#include <X11/X.h>
#define  NEED_REPLIES      /* Tell the header to include the replies */
#define    NEED_EVENTS
#include <X11/Xproto.h>
#include <X11/Intrinsic.h>

#include "memory.h"
#include "network.h"
#include "tool.h"
#include "subscriber.h"
#include "connect.h"
#include "protocol.h"
#include "misc.h"
#include "packet.h"
#include "xpacket.h"
/*
 * Begin Murray Addition
 */
#include "subscriber2.h"
#include "tool2.h"
/*
 * End Murray Addition
 */

#define   DEBUGHandleIncomingClientPacket  0&DEBUGON 
#define   DEBUGHandleClientPopUpRequestPacket 0&DEBUGON
#define   DEBUGHandleIncomingServerPacket   0&DEBUGON
#define   DEBUGProcessReplyPacket   0&DEBUGON
#define   DEBUGHandlePortException   0&DEBUGON
#define   DEBUGHandleIncomingViewerPacket   0&DEBUGON
#define   DEBUGGetReply            0&DEBUGON
#define   DEBUGSendReply         0&DEBUGON
#define   DEBUGRWError                  0&DEBUGON
#define   DEBUGHostResize   0&DEBUGON
#define   DEBUGViewerResize   0&DEBUGON

/*
** FSH  handle unique X11 requests w/color priority
*/
#define SPECIAL_REQUEST_PACKET(opcode) (opcode == X_AllocColor || \
               opcode == X_AllocNamedColor || \
               opcode == X_InternAtom)

#define INVALID_REQUEST 123

typedef struct _ReplyTableEntry {
   char   PacketName[30];      /* MIT Standard Packet Name */
   BYTE   HandlingCode;      /* How TranslatePacket is to handle the request */
   int   OffsetToID[10];      /* Table of offsets to simple XIDs */
   void   (*ProcessFunc)();   /* Function to handle any special processing */
} ReplyTableEntry;
extern ReplyTableEntry ReplyTable[];
extern int has_reply[];
extern BYTE VERBOSE;
extern int OutputCount;
extern BYTE JOIN;
extern BYTE SEQNUM;
extern BYTE POPUP;
extern BYTE XCOLOR;
extern BYTE  MulticastFlag[];  /* If TRUE, include Subscriber in Multicast*/
BOOL SpecialQueryFlag = FALSE;
BOOL HostQueryFlag = FALSE;


/* wang_x add */
extern BYTE RESIZE;
/* end wang_x add */

#ifndef WINDOW
#define WINDOW Window
#endif


#ifdef sun3
#define SUN
#endif

#ifdef sun4
#define SUN
#endif

#ifdef dec5000
#define DEC
#endif

#ifdef rs6000
#define IBM
#endif


/*
 *   HandleIncomingClientPacket
 *
 *   Handle a packet inbound from the client process.  Send its
 *   contents (raw) to the X server and distribute it to whoever
 *   happens to be watching.
 *
 *   xyzzy - Need to see who's looking at the tool.  For now we
 *   xyzzy - just ship it wholesale to everybody.
 *
 *   RETURN VALUE:   void
 */
XtInputCallbackProc HandleIncomingClientPacket( WhatTool, Port, InputID )
Tool *WhatTool;
Socket *Port;
XtInputId *InputID;
{
   xReq      Header;
   BYTE      RequestCode;
   BYTE      *Stuff;
   CARD32      RemainingBytes;
   CARD32      RBytes;
   CARD32      AtomValue;
   CARD32      BitMask;
   BOOL      Override_Redirect;
   CARD32      override_redirect_value;
   BOOL      Parent_Root;
   Subscriber   *Who;
   XWindowAttributes  Win_Attr;
   int        BitNumber;
   int        WhereAt = 0;
   int             abs_x,abs_y;
   Window          dummywin;
   PopUpInfo       PopUpXY;
   PopUpInfo       *PopUpReply;
   Window          root;
   Window          parent;
   Window          children;
   unsigned        int     nchildren;
   XColor          ColorCellDef;
   XColor          RgbDbDef;
   CARD32          PixelValue;
   CARD32          NameTag;
   CARD32      CMap;
   char            *ColorName;
   CARD16          NameLength;
   CARD16          Pad;
   CARD16          NamePacketLength;
   char *ArchPacketP;
   CARD16   HowMany;
   CARD16   HowManyS;
   BYTE   *Stuff2;
   int      special_request;   /* FSH 4-SEP-91 handle unique cases */

   /* Reset the toolkit before leaving */
   XtRemoveInput( *InputID );
   WhatTool->InID=XtAppAddInput( GetAppCon(), *Port, XtInputReadMask, HandleIncomingClientPacket, WhatTool );

   /* Check for a bogus request */
   if ( (RequestCode=CheckByte( *Port )) == NothingAvail) {
      return;
   }

   /*---------------------------------------------------------------
**
** Begin Kantsios addition:
*/
   if (WhatTool->seqnum)
      ++(WhatTool->ToolRequestNumber);
   /* 
**  End Kantsios addition
**-------------------------------------------------------------*/
#   if DEBUGHandleIncomingClientPacket
   fprintf( stderr, "HandleIncomingClientPacket(): ----------- INCOMING PACKET [%d]\n",
       RequestCode );
#   endif

   if ( WhatTool->verbose ){
      if ( ((++(WhatTool->InCount) + WhatTool->OutCount) % (10*OutputCount)) == 0){
         printf("\t%d -> %s -> %d\n",
             WhatTool->OutCount, WhatTool->ToolCommand, WhatTool->InCount);
         fflush(stdout);
      }
      else if ((WhatTool->InCount % OutputCount) == 0) write(1,".",2);
   }


   Header.reqType = RequestCode;
   if (RequestCode < 128){
      WhatTool->pending_reply = has_reply[RequestCode];
      WhatTool->server_pending_reply = has_reply[RequestCode];
   } else {
      WhatTool->pending_reply = FALSE;
      WhatTool->server_pending_reply = FALSE;
   }

   /* Read the remainder of the request header and plug it into the header */

   if ((Stuff=ReadData( *Port, sizeof(xReq)-1, BLOCK ))==NULL) {
      RWError( WhatTool);
      return (NULL);
   }

   memcpy( ((BYTE *)(&Header))+1, Stuff, sizeof(xReq)-1 );
   FreeMem( Stuff );

   RemainingBytes = (SWAP_INT( Header.length, WhatTool->ConnectPrefix.byteOrder )-1) * 4;

   if (RemainingBytes) {
      if ((Stuff = ReadData( *Port, RemainingBytes, BLOCK ))==NULL){
         RWError( WhatTool);
         return (NULL);
      }
   }


   if (RequestCode > 127) {
#   if DEBUGHandleIncomingClientPacket
      fprintf( stderr, "HandleIncomingClientPacket(): Got an Extension Request [%d]\n", RequestCode);
#  endif
      /* OR trash the extension */
      /* To allow extension to go through, comment out the next 3 statements */
      if (WhatTool->seqnum){
         FreeMem(&Header);
         FreeMem( Stuff );
         return;
      } else {
         /* replace extension with null request */
         Header.reqType = X_NoOperation; 
         Header.length = SWAP_INT(1, WhatTool->ConnectPrefix.byteOrder );
         RemainingBytes = 0;
         FreeMem( Stuff );
      }
   }
#   if DEBUGHandleIncomingClientPacket
   fprintf( stderr, "HandleIncomingClientPacket(): Got Request [%d] Header + %d bytes, pending_reply [%d]\n",
       Header.reqType, RemainingBytes,  has_reply[RequestCode] );
   if (has_reply[RequestCode])
      fprintf( stderr, "HandleIncomingClientPacket(): Got Request %d (%s)\n", 
          RequestCode, ReplyTable[RequestCode].PacketName);
#   endif

   /* This is to diagnose error in ChangeProperty request */
   /*
   if (Header.reqType == 18){
      fprintf( stderr, "HandleIncomingClientPacket(): Got Request [%d] Header + %d bytes, pending_reply [%d]\n",
          Header.reqType, RemainingBytes,  has_reply[RequestCode] );
      fprintf( stderr," window atom atom %o8x %o8x %o8x\n", *((CARD32 *)Stuff), 
          *((CARD32 *)(Stuff+4)),*((CARD32 *)(Stuff+8)));
   }
*/
   HandleSpecialQuery(WhatTool,Header.reqType);

   PopUpPacket = FALSE;
   if ((WhatTool->popup)&&(Header.reqType == 1)) {

      Override_Redirect = FALSE;
      Parent_Root = FALSE;

      if (*((WINDOW *)(Stuff+4)) == (WINDOW)WhatTool->HostRootInfo.windowId)
         Parent_Root = TRUE;
      else
         Parent_Root = FALSE;

      BitMask = SWAP_LONG(*((CARD32 *)(Stuff+6*4)),WhatTool->ConnectPrefix.byteOrder);
      if (BitMask & (1 << 9)) {
         for ( BitNumber=0; BitNumber < 9; BitNumber++ )
            if ( BitMask & (1 << BitNumber) )
               WhereAt++;
         /* one codition is to check that the override-redirect falg is true */
         override_redirect_value = SWAP_LONG(*((CARD32 *)(Stuff+7*4+WhereAt*4)),
             WhatTool->ConnectPrefix.byteOrder);
         if (override_redirect_value == 1)
            Override_Redirect = TRUE;
         else
            Override_Redirect = FALSE;


#   if DEBUGHandleClientPopUpRequestPacket
         fprintf( stderr,"The value of Override_Redirect is %d\n", Override_Redirect);
#   endif


         /* another condition is that the window parent is the root window */
         if (Override_Redirect && Parent_Root)
            PopUpPacket = TRUE;
         else
            PopUpPacket = FALSE;
#   if DEBUGHandleClientPopUpRequestPacket
         fprintf( stderr,"The value of PopUpPacket is %d\n", PopUpPacket);
#   endif

      }
      if ((!Override_Redirect) && Parent_Root && WhatTool->FindAbsoluteWindow == FALSE) {
         WhatTool->FindAbsoluteWindow = TRUE;
         WhatTool->AbsoluteWindow = *((WINDOW *)Stuff);
#   if DEBUGHandleClientPopUpRequestPacket
         fprintf( stderr,"AbsoluteWindow for the tool is FOUND%%%%%%%%\n");
         fprintf( stderr,"AbsoluteWindowId===%d Its Parent WIndowID===%d\n",*((WINDOW *)Stuff),*((WINDOW
             *)(Stuff+4)));
#   endif
      }
   }
   if ((WhatTool->popup)&&(PopUpPacket == TRUE) && (WhatTool->FindAbsoluteWindow == TRUE)) {

#   if DEBUGHandleClientPopUpRequestPacket
      fprintf( stderr,"absoulute x and y (not swapped) are: %d , %d\n", *((INT16 *)(Stuff+8)), *((INT16 *)(Stuff+
          10)));
      fprintf( stderr,"AbsoluteWindow==%d\n",WhatTool->AbsoluteWindow);
#   endif

      PopUpXY.opCode = 120;
      PopUpXY.AbsoluteWindow = WhatTool->AbsoluteWindow;
      PopUpXY.PopUpX = *((INT16 *)(Stuff+8));
      PopUpXY.PopUpY = *((INT16 *)(Stuff+10));

      MulticastFlag[0]=FALSE;

      for ( Who=FirstSubscriber(); Who != NULL; Who=NextSubscriber() ) {
         if (Who->SubscriberID == 0) {
            MulticastFlag[0]=FALSE;
            continue;
         }

         MulticastByte(HS_X11Packet);

         MulticastByte(WhatTool->ToolID);

        RBytes = SWAP_LONG(sizeof(PopUpInfo), WhatTool->ConnectPrefix.byteOrder );

         MulticastData(&RBytes, sizeof(CARD32) );
         MulticastData(&PopUpXY, sizeof(PopUpInfo) );
         /*
** Begin Kantsios Addition
*/
         if (WhatTool->seqnum)
         {
            CARD16 temp = htons( WhatTool->ToolRequestNumber); 
            MulticastData( &temp, sizeof(CARD16) );
         }
         /*
** End Kantsios Addition
*/
      }

      if (WhatTool->TokenHolder != HOST_ID) {


#   if DEBUGHandleClientPopUpRequestPacket
         fprintf( stderr,"PopUpPacket == TRUE and the HOST do not has the token for WhatTool\n");
#   endif
         Who = FindSubscriberByID(WhatTool->TokenHolder);

         if ((PopUpReply = (PopUpInfo *)ReadData(Who->Connection,sizeof(PopUpInfo),BLOCK))==NULL){
            RWError( WhatTool);
            return (NULL);
         }

#   if DEBUGHandleClientPopUpRequestPacket
         fprintf( stderr,"Got PopUpReply from token holder\n");
         fprintf( stderr,"Host does not have the token, sending x and y for everyone\n");
#   endif


         *((INT16 *)(Stuff+8)) = PopUpReply->PopUpX;
         *((INT16 *)(Stuff+10)) = PopUpReply->PopUpY;
         /*
         *((INT16 *)(Stuff+8)) = SWAP_INT(PopUpReply->PopUpX,WhatTool->ConnectPrefix.byteOrder);
         *((INT16 *)(Stuff+10)) = SWAP_INT(PopUpReply->PopUpY,WhatTool->ConnectPrefix.byteOrder);
*/
      }
      else {

#   if DEBUGHandleClientPopUpRequestPacket
         fprintf( stderr,"Host have the token, modify x and y \n");
#   endif

         XQueryTree(XtDisplay((Widget *)GetTopLevelWidget()),
             SWAP_LONG(WhatTool->AbsoluteWindow,WhatTool->ConnectPrefix.byteOrder),
             &root, &parent, &children, &nchildren);
         XGetWindowAttributes(XtDisplay((Widget *)GetTopLevelWidget()),
             SWAP_LONG(WhatTool->AbsoluteWindow, WhatTool->ConnectPrefix.byteOrder),&Win_Attr);
         if (root != parent) {
            XTranslateCoordinates(XtDisplay((Widget *)GetTopLevelWidget()),
                SWAP_LONG(WhatTool->AbsoluteWindow, WhatTool->ConnectPrefix.byteOrder),
                SWAP_LONG(WhatTool->HostRootInfo.windowId,WhatTool->ConnectPrefix.byteOrder), Win_Attr.x,
                Win_Attr.y, &abs_x, &abs_y, &children);
            Win_Attr.x = abs_x;
            Win_Attr.y = abs_y;
         }


         *((INT16 *)(Stuff+8)) = SWAP_INT(SWAP_INT(*((INT16 *)(Stuff+8)),WhatTool->ConnectPrefix.byteOrder) -
             Win_Attr.x,WhatTool->ConnectPrefix.byteOrder);
         *((INT16 *)(Stuff+10)) = SWAP_INT(SWAP_INT(*((INT16 *)(Stuff+10)),WhatTool->ConnectPrefix.byteOrder) -
             Win_Attr.y,WhatTool->ConnectPrefix.byteOrder);
      }
#   if DEBUGHandleClientPopUpRequestPacket
      fprintf( stderr,"relative x and y (not swapped) are: %d , %d\n", *((INT16 *)(Stuff+8)), *((INT16 *)(Stuff+
          10)));
#   endif
   }
   /*
  ** Mod FSH - isolate special requests 
  */
   if (special_request = SPECIAL_REQUEST_PACKET(Header.reqType)) {

      /*
      ** Isolate and check additional color related stuff 
      */
      if (Header.reqType != X_InternAtom)
         special_request =  (WhatTool->xcolor &&  /* Check and re-asssign */
         (*(Colormap *)Stuff == WhatTool->HostRootInfo.defaultColormap));
      /*
      ** Now do special processing
      */
      if (special_request) { /* Intern or valid color request w/ XCOLOR on */

         HandleClientSpecialRequest(WhatTool,&Header,Stuff);
         HandleSubSpecialRequest(WhatTool,&Header,Stuff);
         /* 
         **  Archiving is dependent upon JOIN and XCOLOR (For color requests)
         */
         if (WhatTool->join)
            HandleJoinSpecialRequest(WhatTool,&Header,Stuff);
      }

   }


   Stuff2 = (BYTE *) GetMem(RemainingBytes);
   memcpy(Stuff2,Stuff,RemainingBytes);
   /* modify the packet to refelect PopUps in this host */

   if ((WhatTool->popup)&&(PopUpPacket == TRUE) && (WhatTool->FindAbsoluteWindow == TRUE)) {

#   if DEBUGHandleClientPopUpRequestPacket
      fprintf( stderr,"modify the packet to refelect PopUps in this host\n");
#   endif

      XQueryTree(XtDisplay((Widget *)GetTopLevelWidget()),
          SWAP_LONG(WhatTool->AbsoluteWindow,WhatTool->ConnectPrefix.byteOrder),
          &root, &parent, &children, &nchildren);
      XGetWindowAttributes(XtDisplay((Widget *)GetTopLevelWidget()),
          SWAP_LONG(WhatTool->AbsoluteWindow, WhatTool->ConnectPrefix.byteOrder),&Win_Attr);
      if (root != parent) {
         XTranslateCoordinates(XtDisplay((Widget *)GetTopLevelWidget()),
             SWAP_LONG(WhatTool->AbsoluteWindow,WhatTool->ConnectPrefix.byteOrder),
             SWAP_LONG(WhatTool->HostRootInfo.windowId,WhatTool->ConnectPrefix.byteOrder), 
             Win_Attr.x, Win_Attr.y, &abs_x, &abs_y, &children);
         Win_Attr.x = abs_x;
         Win_Attr.y = abs_y;
      }

      *((INT16 *)(Stuff+8)) = SWAP_INT(SWAP_INT(*((INT16 *)(Stuff2+8)),WhatTool->ConnectPrefix.byteOrder) +  
          Win_Attr.x,WhatTool->ConnectPrefix.byteOrder);
      *((INT16 *)(Stuff+10)) = SWAP_INT(SWAP_INT(*((INT16 *)(Stuff2+10)),WhatTool->ConnectPrefix.byteOrder) + 
          Win_Attr.y,WhatTool->ConnectPrefix.byteOrder);
#   if DEBUGHandleClientPopUpRequestPacket
      fprintf( stderr,"Sending to Server PopUp Packet\n");
      fprintf( stderr,"absolute x and y (not swapped) are: %d , %d\n", *((INT16 *)(Stuff+8)), *((INT16 *)(Stuff+
          10)));
      fprintf( stderr,"hostRootWindow==%d\n", *((WINDOW *)(Stuff+4)));
#   endif
   }


   if( WhatTool->join && Header.reqType < 128 ){
      /* Archive the packet */
      ArchPacketP = GetMem(sizeof(xReq)+RemainingBytes);
      memcpy(ArchPacketP,(char *) &Header,sizeof(xReq));
      memcpy(ArchPacketP+sizeof(xReq),Stuff2,RemainingBytes);
      ArchivePacket(WhatTool,ArchPacketP,sizeof(xReq)+RemainingBytes);
      FreeMem(ArchPacketP);
   }

   /* Send the whole mess off to the server */





   /*
** Begin Kantsios
*/

   if ((WhatTool->seqnum) && !(SpecialQueryFlag && WhatTool->TokenHolder != HOST_ID)) {
      ++(WhatTool->ServerRequestNumber);

#   if DEBUGHandleIncomingClientPacket
      fprintf( stderr, "(%d, %d ) \n    ", 
          WhatTool->ServerRequestNumber, WhatTool->ToolRequestNumber );
#   endif
      (void) AddToSNTable(WhatTool, WhatTool->ServerRequestNumber, 
          WhatTool->ToolRequestNumber);

      if (SendData( WhatTool->OutConnection, &Header, sizeof(xReq))==ERROR){

#   if DEBUGHandleIncomingClientPacket
         fprintf( stderr, "HandleIncomingClientPacket(): ERROR in sending Request sent to server\n" );
#   endif
         CleanUpTool(WhatTool);
         return;
      }

      if (RemainingBytes) {
         if(SendData( WhatTool->OutConnection, Stuff, RemainingBytes )==ERROR){

#   if DEBUGHandleIncomingClientPacket
            fprintf( stderr, "HandleIncomingClientPacket(): ERROR in sending Request sent to server\n" );
#   endif
            CleanUpTool(WhatTool);
            return;
         }
      }
#   if DEBUGHandleIncomingClientPacket
      fprintf( stderr, "HandleIncomingClientPacket(): Request %d  sent to server\n",
          Header.reqType );
#   endif
   }
   else if (!WhatTool->seqnum) {

      if (SendData( WhatTool->OutConnection, &Header, sizeof(xReq))==ERROR){

#   if DEBUGHandleIncomingClientPacket
         fprintf( stderr, "HandleIncomingClientPacket(): ERROR in sending Request sent to server\n" );
#   endif
         CleanUpTool(WhatTool);
         return;
      }

      if (RemainingBytes) {
         if(SendData( WhatTool->OutConnection, Stuff, RemainingBytes )==ERROR){

#   if DEBUGHandleIncomingClientPacket
            fprintf( stderr, "HandleIncomingClientPacket(): ERROR in sending Request sent to server\n" );
#   endif
            CleanUpTool(WhatTool);
            return;
         }
      }
#   if DEBUGHandleIncomingClientPacket
      fprintf( stderr, "HandleIncomingClientPacket(): Request %d  sent to server\n",
          Header.reqType );
#   endif
   }
   /* Now ship everybody who's a subscriber a copy, in the usual protcol-istic fashion */


   /* Begin Kantsios@ */
   if (!special_request && !(SpecialQueryFlag && WhatTool->TokenHolder == HOST_ID)) {

      MulticastFlag[0]=FALSE;

      for ( Who=FirstSubscriber(); Who != NULL; Who=NextSubscriber() ) {
         if (Who->SubscriberID == 0) {
            MulticastFlag[0]=FALSE;
            continue;
         }
         if( WhatTool->join && ArchiveToBeSent(Who,WhatTool) ){
            MulticastFlag[Who->SubscriberID]=FALSE;
            continue;
         }
      }
#       if DEBUGHandleIncomingClientPacket
      fprintf( stderr, "HandleIncomingClientPacket(): ... Multicasting X11Packet ...  \n");
#      endif

      MulticastByte(HS_X11Packet);
      MulticastByte(WhatTool->ToolID);

  RBytes = SWAP_LONG( RemainingBytes+sizeof(xReq),
       WhatTool->ConnectPrefix.byteOrder );

      MulticastData(&RBytes, sizeof(CARD32) );
      MulticastData(&Header, sizeof(xReq));
      if (RemainingBytes)
         MulticastData(Stuff2, RemainingBytes );
      if (WhatTool->seqnum)
      {
         CARD16 temp = htons( WhatTool->ToolRequestNumber); 
         MulticastData( &temp, sizeof(CARD16) );
      }
      /*
** End Kantsios Addition
*/
      {
         int  i;
         for ( i=0; i<256; i++ )
            MulticastFlag[i] = TRUE;
      }

   }
   if ( RemainingBytes )
      FreeMem( Stuff );
   FreeMem( Stuff2 );

#       if DEBUGHandleIncomingClientPacket
   fprintf( stderr, "HandleIncomingClientPacket(): ... FINISH handling X11Packet ...  \n");
#      endif

}




/*
 *   BYTE *GetReply
 *
 *   Check to see if the server has anything to send us.
 *
 *   RETURN VALUE:   Pointer to sent data if and is available
 *         NULL otherwise
 */
BYTE *GetReply( Source, WhatTool )
Socket *Source;
Tool *WhatTool;
{
   BYTE      *First32;
   BYTE      *Remainder;
   int      ReplyLength;
   xError     *ErrorPacket;
   xEvent *EventPacket;
   CARD16 SQNum;

   if ( (First32=ReadData( Source, sizeof(xGenericReply), BLOCK)) == NULL ) {
      RWError( WhatTool);
      return (NULL);
   } else{
    SQNum = ((xGenericReply *)First32)->sequenceNumber;
    SQNum = SWAP_INT(SQNum, WhatTool->ConnectPrefix.byteOrder );
#   if DEBUGGetReply
      fprintf(stderr, "DEBUGGetReply(): Got First32 of REPLY/EVENT\n" );
      fprintf(stderr,"FRESH...Sequence No. is: %04d \n", SQNum);
# endif
   /*---------------------------------------------------------------
**
** Begin Kantsios addition:
*/
   if (WhatTool->seqnum) 
      SQNum = (CARD16) GetClientSeqNum( WhatTool, SQNum);
#   if DEBUGGetReply
      fprintf(stderr,"MAPPED...Sequence No. is: %04d\n", SQNum);
# endif

      ((xGenericReply *)(First32))->sequenceNumber = SWAP_INT(SQNum, 
				WhatTool->ConnectPrefix.byteOrder );


      if ( *First32 == 1 ) {
         ReplyLength = (SWAP_LONG( ((xGenericReply *)First32)->length, 
                        WhatTool->ConnectPrefix.byteOrder)) * 4;
#   if DEBUGGetReply
         fprintf(stderr, "DEBUGGetReply(): Detected a REPLY, length is %d Bytes\n", 
             32 + ReplyLength);
# endif
         if ( ReplyLength == 0 ) {
            return( First32 );
         }
         else {
#   if DEBUGGetReply
            fprintf( stderr,"DEBUGGetReply(): reading %d remainder of REPLY\n",
                ReplyLength );
#   endif

            if ((Remainder = ReadData( Source, ReplyLength, BLOCK ))==NULL){
               RWError( WhatTool);
               return (NULL);
            }
#   if DEBUGGetReply
            fprintf( stderr,"DEBUGGetReply(): finish reading remainder of REPLY\n" );
#   endif
            First32 = (BYTE *)realloc( First32, 32+ReplyLength );
            memcpy( First32+32, Remainder, ReplyLength );
            free( Remainder );
            return( First32 );
         }

      }

      /* Anything else must be an event or an error... Always 32 bytes */

      ErrorPacket = (xError *)First32;
      if (ErrorPacket->type == 0){

#   if DEBUGGetReply
         fprintf( stderr,"DEBUGGetReply():Detected an ERROR, errCode (%d), sequenceNumber(%d) \n",
             ErrorPacket-> errorCode, ErrorPacket->sequenceNumber);
#   endif

         if ((ErrorPacket-> errorCode == 1) && (ErrorPacket->majorCode == INVALID_REQUEST) && 
             (WhatTool->seqMap.MakeRoomFlag == TRUE)) {
#   if DEBUGGetReply
            fprintf(stderr,"ISA REPLY TO FLUSHING...................\n");
            fprintf(stderr," MakeRoomSeqNum: %d\n", WhatTool->seqMap.MakeRoomSeqNum);
#   endif
            WhatTool->seqMap.MakeRoomFlag = FALSE;
            return(NULL);
         }

      } else {
         EventPacket = (xEvent *)First32;
#   if DEBUGGetReply
         fprintf( stderr,"DEBUGGetReply():Detected an EVENT, OpCode (%d), sequenceNumber(%d)\n", 
             EventPacket->u.u.type, EventPacket->u.u.sequenceNumber);
#   endif

      }
      return( First32 );
   }
}


/*
 *   void SendReply
 *
 *   Send a reply, event, or error.  Detemine which it is by examining the
 *   first byte in the packet and act accordingly.
 *
 *   RETURN VALUE:   (void)
 */

SendReply( Dest, Packet, WhatTool )
Socket Dest;
BYTE *Packet;
Tool *WhatTool;

{
   int   Length;
   int   ReplyLength;


#   if DEBUGSendReply
   fprintf( stderr,"DEBUGSendReply(): Sendig  a Reply \n");
#   endif

   if (GetMode() ){
#   if DEBUGSendReply
      fprintf( stderr,"DEBUGSendReply(): Sequence Number, UnModified %04x \n", 
          ((((xGenericReply *)Packet)->sequenceNumber)));
      fprintf( stderr,"DEBUGSendReply(): Tool Last Sequence %04x \n", 
          (WhatTool->LastSequenceNumber));
      fprintf( stderr,"DEBUGSendReply(): Sequence Number, Modified %04x \n", 
          ((((xGenericReply *)Packet)->sequenceNumber)));
#   endif
   }

   if ( *Packet == 1 ) {
      if ( GetMode() ){
         WhatTool->server_pending_reply = FALSE;
         WhatTool->pending_reply = FALSE;
      }

      ReplyLength = (SWAP_LONG( ((xGenericReply *)Packet)->length, WhatTool->ConnectPrefix.byteOrder)) * 4;
      Length = 32 +  ReplyLength ;
#   if DEBUGSendReply
      fprintf( stderr,"DEBUGSendReply(): Reply Length = %d Bytes\n", Length );
#   endif

   } else {
      Length = 32;
#   if DEBUGSendReply
      fprintf( stderr,"DEBUGSendReply(): Event Length = %d Bytes\n", Length );
#   endif

   }
   /*-------------------------------------------------------------
** Begin Kantsios Addition
*/
   if ( GetMode() ){
      if (((xGenericReply *)(Packet))->sequenceNumber < WhatTool->LastSequenceNumber)
         ((xGenericReply *)(Packet))->sequenceNumber = WhatTool->LastSequenceNumber;
      if (((xGenericReply *)(Packet))->sequenceNumber > WhatTool->LastSequenceNumber)
         WhatTool->LastSequenceNumber = ((xGenericReply *)(Packet))->sequenceNumber;

   }

   /*
**  End Kantsios Addition
-------------------------------------------------------------*/

   if (SendData( Dest, Packet, Length ) == ERROR ) {
#   if DEBUGSendReply
      fprintf( stderr,"DEBUGSendReply(): Could't send data, Tool %s PID %d is Terminated\n",

          WhatTool->ToolCommand, WhatTool->ToolPID);
#   endif
      if (GetMode()) {
         CleanUpTool(WhatTool);
         return(ERROR);
      }
      else EndSession();
#   if DEBUGSendReply
      fprintf( stderr,"DEBUGSendReply(): Cleaned Up\n");
#   endif
   } else {

#   if DEBUGSendReply
      fprintf( stderr,"DEBUGSendReply(): Completed.. \n");
      fflush (stderr);
#   endif
   }

}



/*
 *   HandleIncomingServerPacket
 *
 *   Handle a packet outbound from the server and return it
 *   to the client.
 *
 *   RETURN VALUE:   void
 */
XtInputCallbackProc HandleIncomingServerPacket( WhatTool, Port, InputID )
Tool *WhatTool;
Socket *Port;
XtInputId *InputID;
{
   BYTE      *Packet;
   /* Reset the toolkit before leaving */
   XtRemoveInput( *InputID );
   WhatTool->OutID=XtAppAddInput( GetAppCon(), *Port, XtInputReadMask, HandleIncomingServerPacket, WhatTool );
#   if DEBUGHandleIncomingServerPacket
   fprintf( stderr, "HandleIncomingServerPacket(): event/reply  detected\n" );
#   endif

   if ((Packet=GetReply(*Port, WhatTool)) == NULL) {
      FreeMem( Packet );
      return;
   }

   ProcessReplyPacket(WhatTool, Packet);
}

ProcessReplyPacket(WhatTool, Packet)
Tool *WhatTool;
BYTE      *Packet;
{
   /* wang_x add */
   int        PreWidth, PreHeight;
   Subscriber  *Who;
   ResizeInfo   Res;
   ResizeInfo   OutPack;
   Window   re_window;
   INT16     re_width;
   INT16     re_height;

   /* end wang_x add */


   if ( WhatTool->verbose ){
      if ( ((++(WhatTool->OutCount) + WhatTool->InCount) % (10*OutputCount)) == 0){
         printf("\t%d -> %s -> %d\n",
             WhatTool->OutCount, WhatTool->ToolCommand, WhatTool->InCount);
         fflush(stdout);
      }
      else if ((WhatTool->OutCount % OutputCount) == 0)  write(1,"|", 2);
   }
   /* wang_x add */
   if (*Packet == 22 && WhatTool->resize) {
#   if DEBUGHostResize
      fprintf( stderr, "HostResize: Resize packet detected\n" );
#   endif

      Res.re_window =  ((xEvent *)Packet)->u.configureNotify.window;
      Res.re_width = ((xEvent *)Packet)->u.configureNotify.width;
      Res.re_height = ((xEvent *)Packet)->u.configureNotify.height;
#   if DEBUGHostResize
   re_window = SWAP_LONG(Res.re_window, WhatTool->ConnectPrefix.byteOrder);
   re_width = SWAP_INT(Res.re_width,WhatTool->ConnectPrefix.byteOrder);
   re_height = SWAP_INT(Res.re_height, WhatTool->ConnectPrefix.byteOrder);
      fprintf( stderr, "HostResize: Sending Resize pckt to parts 0x%08x  %dx%d\n",
			re_window, re_width, re_height );
#   endif
      MulticastFlag[0]=FALSE;
      MulticastByte(HS_ResizePacket);
      MulticastByte(WhatTool->ToolID);
      bcopy(&Res, &OutPack, sizeof(ResizeInfo));
      MulticastData(&OutPack, sizeof(Res) );
      MulticastFlag[0]=TRUE;

   }
   /* end wang_x add */

   if (*Packet == 12){
      /* Exposure event, send it out */
#   if DEBUGProcessReplyPacket
      fprintf( stderr, "ProcessReplyPacket(): Exposure event sent to client\n" );
#   endif
      SendReply( WhatTool->InConnection, Packet, WhatTool);
      FreeMem( Packet );
      return;
   }

   if ((*Packet == 1) || (*Packet == 0)){
      if ( (WhatTool->TokenHolder == HOST_ID ) && (SpecialQueryFlag == TRUE)){
         SpecialQueryFlag = FALSE;
#   if DEBUGProcessReplyPacket
         if (*Packet == 1)
            fprintf( stderr, "ProcessReplyPacket(): SpecialQuery:REPLY is sent\n");
         else
            fprintf( stderr, "ProcessReplyPacket(): SpecialQuery:ERROR is sent\n");
#   endif

         WhatTool->server_pending_reply = FALSE;
         WhatTool->pending_reply = FALSE;
         SendReply( WhatTool->InConnection, Packet, WhatTool);
         FreeMem( Packet );
         return;
      }
      if (HostQueryFlag == TRUE){
         HostQueryFlag = FALSE;
#   if DEBUGProcessReplyPacket
         if (*Packet == 1)
            fprintf( stderr, "ProcessReplyPacket(): HostQuery:reply is sent\n");
         else
            fprintf( stderr, "ProcessReplyPacket(): HostQuery:ERROR is sent\n");
#   endif
         WhatTool->server_pending_reply = FALSE;
         WhatTool->pending_reply = FALSE;
         SendReply( WhatTool->InConnection, Packet, WhatTool);
         FreeMem( Packet );
         return;
      }
      if  (WhatTool->TokenHolder == HOST_ID ) {
#   if DEBUGProcessReplyPacket
         if (*Packet == 1)
            fprintf( stderr, "ProcessReplyPacket(): :STRANGE reply is sent\n");
         else
            fprintf( stderr, "ProcessReplyPacket(): :ERROR is sent\n");
#   endif
         SendReply( WhatTool->InConnection, Packet, WhatTool);
         FreeMem( Packet );
         return;
      } else{
#   if DEBUGProcessReplyPacket
         if (*Packet == 1)
            fprintf( stderr, "ProcessReplyPacket(): :TRASH REPLY (no token) \n");
         else
            fprintf( stderr, "ProcessReplyPacket(): :TRASH ERROR (no token) \n");
#   endif
         FreeMem( Packet );
         return;
      }
   } else {
      if ( DetermineHostRouting( WhatTool, Packet) ) {
#         if DEBUGProcessReplyPacket
         fprintf( stderr, "ProcessReplyPacket(): Sending EVENT to Host\n" );
#         endif

         SendReply( WhatTool->InConnection, Packet, WhatTool);
#   if DEBUGProcessReplyPacket
         fprintf( stderr, "ProcessReplyPacket(): Host: Event is sent\n");
#   endif
      }
   }
   FreeMem( Packet );
}


/*
 *   HandlePortException
 *
 *   Handle any I/O exeptions from the input or output port(s) by
 *   closing 'em up and removing the tool from the list.
 *
 *   RETURN VALUE:   void
 */
XtInputCallbackProc HandlePortException( WhatTool, Port, InputID )
Tool *WhatTool;
Socket *Port;
XtInputId *InputID;
{
   /* Tell Xt to forget about us and close up the ports */
   shutdown( WhatTool->InConnection, 2 );
   XtRemoveInput( WhatTool->InID );
   XtRemoveInput( WhatTool->OutID );
   XtRemoveInput( WhatTool->InExceptID );
   XtRemoveInput( WhatTool->OutExceptID );


#   if DEBUGHandlePortException
   fprintf( stderr, "HandlePortException(): Port exception on %dConnection\n",
       ( *Port == WhatTool->InConnection ? "In" : "Out" ) );
#   endif

   UpdateMessage( "Tool %s has %s.\n", ( *Port == WhatTool->InConnection ? "terminated and disconnected" : "lost its connection to the server" ) );

   /* Cook the tool */

   DeleteTool( WhatTool );

}


/*
 *   HandleIncomingViewerPacket
 *
 *   Handle an inbound event from a viewer's server.  Do all
 *   translations, and ship it, along with the appropriate protocol
 *   stuff to the host if needed.
 *
 *   RETURN VALUE:   void
 */
XtInputCallbackProc HandleIncomingViewerPacket( WhatTool, Port, InputID )
Tool *WhatTool;
Socket *Port;
XtInputId *InputID;
{
   BYTE      *Packet;
   /* wang_x add */
   int       lwbase,hwbase,rw;
   short       width,height;
   ResizeInfo     Res;
   ResizeInfo     OutPack;
   /* end wang_x add */


   /* Reset the toolkit before leaving */
   XtRemoveInput( *InputID );
   WhatTool->OutID=XtAppAddInput( GetAppCon(), *Port, XtInputReadMask, HandleIncomingViewerPacket, WhatTool
       );
#   if DEBUGHandleIncomingViewerPacket
   fprintf( stderr, "HandleIncomingViewerPacket(): REPLY/EVENT  detected\n" );
#   endif

   if ((Packet=GetReply(*Port, WhatTool)) == NULL) {
      FreeMem( Packet );
      return;
   }

   if ( WhatTool->verbose ){
      if ( ((++(WhatTool->OutCount) + WhatTool->InCount) % (10*OutputCount)) == 0){
         printf("\t%d -> %s -> %d\n",
             WhatTool->OutCount, WhatTool->ToolCommand, WhatTool->InCount);
         fflush(stdout);
      }
      else if ((WhatTool->InCount % OutputCount) == 0) write(1,"|",2);
   }
   /* wang_x add */
   if (*Packet == 22 && WhatTool->resize) {
      rw = ((xEvent *)Packet)->u.configureNotify.window;
      Res.re_window = TranslateBackward( WhatTool, rw, S_XLATE );
      Res.re_width = ((xEvent *)Packet)->u.configureNotify.width;
      Res.re_height = ((xEvent *)Packet)->u.configureNotify.height;
#   if DEBUGViewerResize
      fprintf( stderr, "ViewerResize: Resize packet detected\n" );
#   endif

      if (WhatTool->TokenHolder == GetLocalSubID() ){
#   if DEBUGViewerResize
         fprintf( stderr, "ViewerResize: Have Token, send to Host Resize Info\n" );
#   endif
         Res.corr_me = 0;
      }
      else
      {
#   if DEBUGViewerResize
         fprintf( stderr, "ViewerResize: Does Not Have Token, send to Host Resize Info\n" );
#   endif
         Res.corr_me = 1;
      }
         SendByte( GetHostConnection(), SH_ResizePacket)  ;
         SendByte( GetHostConnection(), WhatTool-> ToolID) ;
         bcopy(&Res, &OutPack, sizeof(ResizeInfo));
         SendData( GetHostConnection(), &OutPack, sizeof(Res) ) ;
   }

   /* end wang_x add */

   if (*Packet == 12){
      TranslateViewerEvent( Packet, WhatTool);
      if (SendByte( GetHostConnection(), SH_Event )== ERROR) EndSession();
      if (SendByte( GetHostConnection(), WhatTool->ToolID )== ERROR) EndSession;
      SendReply( GetHostConnection(), Packet, WhatTool);

#   if DEBUGHandleIncomingViewerPacket
      fprintf( stderr, "HandleIncomingViewerPacket(): Exposure event sent to host\n" );
#   endif
      FreeMem( Packet );
      return;
   }
   if ((*Packet == 1) || (*Packet == 0)){
      if ( (WhatTool->TokenHolder == GetLocalSubID() ) && (SpecialQueryFlag == TRUE)){
         SpecialQueryFlag = FALSE;
#   if DEBUGHandleIncomingViewerPacket
         if (*Packet == 1)
            fprintf( stderr, "HandleIncomingViewerPacket(): SpecialQuery:reply is sent\n");
         else
            fprintf( stderr, "HandleIncomingViewerPacket(): SpecialQuery:ERROR is sent\n");

#   endif

         TranslateReply( Packet, WhatTool);
         if (SendByte( GetHostConnection(), SH_Reply )==ERROR) EndSession();
         if(SendByte( GetHostConnection(), WhatTool->ToolID )==ERROR) EndSession();
         SendReply( GetHostConnection(), Packet, WhatTool);
         FreeMem( Packet );
         return;
      }
      if (HostQueryFlag == TRUE){
         HostQueryFlag = FALSE;
#   if DEBUGHandleIncomingViewerPacket
         if (*Packet == 1)
            fprintf( stderr, "HandleIncomingViewerPacket(): HostQuery:reply is trashed\n");
         else
            fprintf( stderr, "HandleIncomingViewerPacket(): HostQuery:ERROR is trashed\n");
#   endif
         FreeMem( Packet );
         return;
      }
      if ((SpecialQueryFlag == TRUE) && (WhatTool->TokenHolder != GetLocalSubID())){
         SpecialQueryFlag = FALSE;
#   if DEBUGHandleIncomingViewerPacket
         if (*Packet == 1)
            fprintf( stderr, "HandleIncomingViewerPacket(): SpecialQuery:reply is TRASHED (no token)\n");
         else
            fprintf( stderr, "HandleIncomingViewerPacket(): SpecialQuery:ERROR is TRASHED (no token)\n");
#   endif
         FreeMem( Packet );
         return;
      }
      if ( WhatTool->TokenHolder == GetLocalSubID() ){
#   if DEBUGHandleIncomingViewerPacket
         if (*Packet == 1)
            fprintf( stderr, "HandleIncomingViewerPacket(): :STARNGE reply is sent\n");
         else
            fprintf( stderr, "HandleIncomingViewerPacket(): :ERROR is sent\n");
#   endif

         TranslateReply( Packet, WhatTool);
         if (SendByte( GetHostConnection(), SH_Reply )==ERROR) EndSession();
         if(SendByte( GetHostConnection(), WhatTool->ToolID )==ERROR) EndSession();
         SendReply( GetHostConnection(), Packet, WhatTool);
         FreeMem( Packet );
         return;
      } else{
#   if DEBUGHandleIncomingViewerPacket
         if (*Packet == 1)
            fprintf( stderr, "HandleIncomingViewerPacket(): :TRASH reply (no token) \n");
         else
            fprintf( stderr, "HandleIncomingViewerPacket(): :TRASH error (no token) \n");
#   endif
         FreeMem( Packet );
         return;
      }
   } else {
#   if DEBUGHandleIncomingViewerPacket
      fprintf( stderr, "HandleIncomingViewerPacket(): Event Packet\n" );
#   endif

      if ( DetermineViewerRouting( WhatTool, Packet, GetLocalSubID() ) ) {
#         if DEBUGHandleIncomingViewerPacket
         fprintf( stderr, "HandleIncomingViewerPacket(): Sending EVENT to Host\n" );
#         endif
         TranslateViewerEvent( Packet, WhatTool ) ;
         if (SendByte( GetHostConnection(), SH_Event )==ERROR) EndSession();
         if (SendByte( GetHostConnection(), WhatTool->ToolID )==ERROR) EndSession();
         SendReply( GetHostConnection(), Packet, WhatTool);

      }
   }

   FreeMem( Packet );

}

RWError(WhatTool)
Tool *WhatTool;
{
#   if DEBUGRWError
   fprintf(stderr, "DEBUGRWError(): Got ERROR \n");
   fprintf(stderr, "DEBUGRWError(): terminating tool %d\n", WhatTool->ToolID);
# endif
   if (GetMode()){
      CleanUpTool(WhatTool);
      return (ERROR);
   }
   else {
      if ( SendByte(GetHostConnection(), SH_StopTool)==ERROR) EndSession();
      if ( SendByte(GetHostConnection(), WhatTool->ToolID)==ERROR) EndSession();
      CleanUpTool(WhatTool);
      return(ERROR);
   }
}

SendMakeRoomRequest(WhatTool)
Tool *WhatTool;
{
   xReq Header;
   Header.reqType = INVALID_REQUEST; /* illigal opcode */
   /*
   Header.reqType = X_ListHosts;
   Header.reqType = X_NoOperation;
*/
   Header.length = SWAP_INT(1, WhatTool->ConnectPrefix.byteOrder );
   SendData( WhatTool->OutConnection, &Header, sizeof(xReq));
   /*
fprintf(stderr,"sent flush request, seqNum %4d\n",WhatTool->seqMap.MakeRoomSeqNum);
*/
}

TrashPacket(WhatTool, Port)
Tool *WhatTool;
Socket *Port;
{
   xReq      Head;
   BYTE      *Tail;
   CARD32      RemainingBytes;

   fprintf(stderr,"Trashing..........\n");
   if ((Tail=ReadData( *Port, sizeof(xReq)-1, BLOCK ))==NULL) {
      RWError( WhatTool);
      return (NULL);
   }
   memcpy( ((BYTE *)(&Head))+1, Tail, sizeof(xReq)-1 );
   FreeMem( Tail );

   RemainingBytes = (SWAP_INT( Head.length, WhatTool->ConnectPrefix.byteOrder )-1) * 4
       ;
   if (RemainingBytes) {
      if ((Tail = ReadData( *Port, RemainingBytes, BLOCK ))==NULL){
         RWError( WhatTool);
         return (NULL);
      }
   }

   FreeMem(&Head);
   FreeMem( Tail );
}
