/*
 * Copyright (c) 1991, 1993 Old Dominion University.
 * Copyright (c) 1991, 1993 University of North Carolina at Chapel Hill.
 * All rights reserved.
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  Old Dominion University and the University of North
 * Carolina at Chapel Hill make no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 */


/*
** The following is a datatype circular queue of sequence numbers
** the queue is used to take care of the sequence number problem within
** XTV.  Two of the operators given will not be needed by XTV, unless
** they are used for some type of debuggin purpose, or theu could be
** used to help determine the best size for the circular queue.
** The following was written by:
**		George Kantsios		kants_g@cs.odu.edu
*/


#define   DEBUGSeqQue  0&DEBUGON



/*
** The circluar sequence map table has two integer
** pointers defined in the structure, in the beginning
** I want these guys both pointing to the zeroth entry.
*/

#include "tool.h"

int WaterMark;

int IntCircSeqMap( Map )
   CircSeqMap *Map;
{
        int i;
	int Percentage;

	Percentage = 0.2 * MAX_SEQ_TABLE;
	WaterMark =  (Percentage < 2 ? 2 : Percentage);
	Map -> head = 0;
	Map -> tail = 0;
	Map -> first = TRUE;
        Map -> MakeRoomFlag = FALSE;
for(i=0;i++;i<MAX_SEQ_TABLE){
       Map ->table[i].Server_SN = 0;
       Map ->table[i].Tool_SN   =0;
}
#   if DEBUGSeqQue
fprintf(stderr,"WaterMark: %d\n", WaterMark);
#   endif

}





/*
** This function returns the integer pointer to be used as the
** pointer to a new entry in the circular sequence number table.
** Notice tail of the cicular sequence map is changed.  Basically,
** this is allocating space on my home brew pointer heap.
*/

int nextEntry(Map)
  CircSeqMap *Map;
{
   if (Map->first)
      return( Map->tail );
   else
      return( (Map->tail = (Map->tail+1) % MAX_SEQ_TABLE) );
}






/* 
** This function adds a sequence number and it's coorsponding
** mapping into the CircSeqMap, adjusting the head and tail 
** accordingly.
*/

AddToSNTable(WhatTool, ServerSN, ToolSN)
   Tool *WhatTool;
   CARD16 ServerSN, ToolSN;
{
   	CircSeqMap *Map;
	int next;

#   if DEBUGSeqQue
          fprintf(stderr, "Inserting seq numbers %d %d\n", ServerSN, ToolSN);
#  endif

   	Map = &(WhatTool->seqMap);
	next = nextEntry(Map);
	
	if ( ((next+WaterMark)% MAX_SEQ_TABLE == Map->head) && !(Map->first) &&
		!(Map->MakeRoomFlag) ) {
#   if DEBUGSeqQue
	   fprintf(stderr, "Circular Queue exceeding maximum limit\n");
           fprintf(stderr, "next: %d, WaterMark: %d, head: %d\n", 
           next, WaterMark, Map->head);
	   fprintf(stderr, "Sending flush packet\n");
#  endif
	   Map -> table[next].Server_SN     = ServerSN;
	   Map -> table[next].Tool_SN     = ToolSN;
	   next = nextEntry(Map);
	   Map->MakeRoomSeqNum = ++(WhatTool->ServerRequestNumber);
	   Map->MakeRoomFlag = TRUE;
	   Map -> table[next].Server_SN     = Map->MakeRoomSeqNum;
	   Map -> table[next].Tool_SN     = ERROR;
	   SendMakeRoomRequest(WhatTool);
	} 
        else if ( (next == Map->head) && !(Map->first) && (Map->MakeRoomFlag)){
		fprintf(stderr, "Fatal Error: Circular Queue Probelm\n");
		return(ERROR);
            }
	else {
	   Map -> table[next].Server_SN	= ServerSN;
	   Map -> table[next].Tool_SN	= ToolSN;
	   Map->first = FALSE;
	}
}

/* 
** This function locates a sequence number in the table and returns
** the coorsponding mapping.
*/

CARD16 GetClientSeqNum(WhatTool, ServerRequestNumber )
   Tool *WhatTool;
   CARD16  ServerRequestNumber;
{
   int theTop;
   CircSeqMap *Map;
        Map = &(WhatTool->seqMap);
	
	/* 
	** Search for ServerRequestNumber in the origin part of Map.Table
	** and return the corresponding mapping.
        ** Note that numbers in que are sorted, thus we ca use < operator
	** rather than = operator in searchin (fix wahab).	
	** That greatly improves performance
	*/
#   if DEBUGSeqQue
           fprintf(stderr,"looking for %d \n", ServerRequestNumber);
#   endif

	theTop = Map->head;
	while( (Map->table[theTop].Server_SN) < ServerRequestNumber ) {
	   if (theTop == Map->tail){
	      /*
	      ** If we fail to find it in the map, return 0
	      */
#   if DEBUGSeqQue
           fprintf(stderr,"fail to mapp seq number\n");
#   endif
	      return ( (CARD16) 0 );
	      
	   }
	   theTop = (theTop+1) % MAX_SEQ_TABLE;
	}
	/* it reaches this point only if it's a match or > */

	Map->head = theTop;

	if( (Map->table[theTop].Server_SN) == ServerRequestNumber ) { 
	
#   if DEBUGSeqQue
           fprintf(stderr,"found seq number %d\n", Map->table[theTop].Tool_SN);
#   endif

		return( Map->table[theTop].Tool_SN );
        }else{
#   if DEBUGSeqQue
           fprintf(stderr,"fail to mapp seq number\n");
#   endif

	      return ( (CARD16) 0 );
}
}






/*
**  For testing purposes (not needed) I have included the following function that is used
**  to display all the values within the sequence table map.
*/
void ShowCircSeqMap( Map )
   CircSeqMap Map;
{
   if (!Map.first) {
      if ( Map.head != Map.tail ) {
         printf("[ (%d %d)", Map.table[Map.head].Server_SN, Map.table[Map.head].Tool_SN );
         Map.head = (Map.head+1) % MAX_SEQ_TABLE;
         while (Map.head != Map.tail){
            printf(", (%d %d)", Map.table[Map.head].Server_SN, Map.table[Map.head].Tool_SN );
            Map.head = (Map.head+1) % MAX_SEQ_TABLE;
         }
         printf(", (%d %d) ]\n", Map.table[Map.head].Server_SN, Map.table[Map.head].Tool_SN );
      }
      else
	 printf("[ (%d %d) ]\n", Map.table[Map.head].Server_SN, Map.table[Map.head].Tool_SN );
   }
   else
      printf("[-empty-]\n");
}



/*
**---------------------End of the file seq.c------------------------
*/
